# -*- coding: utf-8 -*-
##############################################################################
#    Daniel Reis
#    
#    OpenERP, Open Source Management Solution
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import time
import pooler
#import tools
#from tools.translate import _
from tools.safe_eval import safe_eval
from osv import fields, osv

import logging
_logger = logging.getLogger(__name__)

#TODO: implement this as an object, so it can be safely be accessed by "changed('state')" instead of "changed.get('state')"
DEFAULT_EVALDICT = {
    'old': {},
    'new': {},
    'changed': {},
    'creating': False,
    'inserting': False,
    'writing': False,
    'updating': False,
}
###############################################################################
# Base Action Rule
# (extends base_action_rule.py and crm_action_rule.py)
#
#   * format_mail() allows using additional fields in the email body:
#        ref, using %(object_ref)s
#        translated state name, using %(object_state)s
#        uppercase translated state name, using %(object_STATE)s
#   * email_send() allows a custom subject line, that should be provided 
#     in the first line of the "body", using the prefix "Subject:".
#
###############################################################################

class base_action_rule(osv.osv):
    _inherit = 'base.action.rule'
    _columns = {
        'trg_evalexpr': fields.text('Evaluated expression',\
            help="""Python expression, able to use a "new" and "old" dictionaries, with the changed columns."""), 
#        'trg_evalexpr_debug': fields.boolean('Debug Evaluated expression',\
#            help="""Write detailed information to log, to help debugging trigger expressions."""), 
        'email_template_id': fields.many2one('email.template', 'E-mail template',
            domain="[('model_id','=',model_id)]" ),
        'email_template_force': fields.boolean('Send immediately',
            help='If not checked, it will be sent the next time the e-mail scheduler runs.'),
    }

    def _check_evalexpr(self, cr, uid, ids, context=None):
         action = self.browse(cr, uid, ids[0], context=context)
         if action.trg_evalexpr:
             #Test if expression is valid, against an empty dict
             #Newlines are tolerated
             safe_eval(action.trg_evalexpr.replace('\n',' '), {}, DEFAULT_EVALDICT) 
         return True

    _constraints = [
        (_check_evalexpr, 'Error: Your evaluated expression is not valid!', ['trg_evalexpr'])
    ]

    def _create(self, old_create, model, context=None):
        """
        Return a wrapper around `old_create` calling both `old_create` and
        `post_action`, in that order.
        """
        def wrapper(cr, uid, vals, context=context):
            if context is None:
                context = {}
            #store new and old values in context, to be used in trigger expressions
            context.update({
                '_action_old': {},
                '_action_new': vals,
                })
            new_id = old_create(cr, uid, vals, context=context)
            #rint 'WRAPPER_CREATE: new=', vals #deepdebug
            if not context.get('action'):
                context.update( {'_action_trigger': 'create'} ) #activate trigger
                self.post_action(cr, uid, [new_id], model, context=context)
            return new_id
        return wrapper
    
    def _write(self, old_write, model, context=None):
        """
        Return a wrapper around `old_write` calling both `old_write` and
        `post_action`, in that order.
        """
        def wrapper(cr, uid, ids, vals, context=context):
            if context is None:
               context = {}
            if isinstance(ids, (str, int, long)):
                ids = [ids]
            #store new and old values in context, to be used in trigger expressions
            olds = self.pool.get(model).read(cr, uid, ids, context=context)
            context.update( {
                '_action_old': olds,
                '_action_new': vals,
                })
            #rint 'WRAPPER_CREATE: new=', vals #deepdebug
            old_write(cr, uid, ids, vals, context=context)
            if not context.get('action'):
                context.update( {'_action_trigger': 'write'} ) #activate trigger
                self.post_action(cr, uid, ids, model, context=context)
            return True
        return wrapper

    #Actions can be triggered from evaluated expression, using values from dictionaries 'old' and 'new'
    # new - is the dictionay passed to the create/write method, therefor it contains only the fields to write.
    # old - on create is an empty dict {} or None value; on write constains the values before the write is executed, given by the read() method
    #Example, check if responsible changed: not old and old['user_id']!=new['user_id']
    def do_check(self, cr, uid, action, obj, context={}):
        ok = super(base_action_rule, self).do_check(cr, uid, action, obj, context=context)
        if ok and action.trg_evalexpr:
            ok = False
            if context.get('_action_trigger'):
                # is_ins flag: signals inserting (create) or updating (write)
                is_ins = context.get('_action_trigger') == 'create'
                #rint '....is_ins = ', is_ins
                #If no changed values, exit with False
                #Abort if called from crm_case._action, to duplicate triggering
                if not context.get('_action_new') or context.get('state_to'):
                    return False
                # old dict: holds original values for all columns (before the write/update)
                old = {}
                for x in context.get('_action_old'): 
                    if x.get('id') == obj.id: #Old is a list of records
                        old = x
                        break
                #Normalize tuples (id, name) on "old" into id only form
                for x in old:
                    if isinstance(old[x], tuple):
                        old[x] = old[x][0]
                # changed dict: holds only the changed values; available only on write/update
                changed = {}
                if not is_ins:
                    for k, v in context.get('_action_new').items():
                        #rint '\t', k, ':', v, '<=', old.get(k)
                        if old.get(k) != v:
                            changed.update( {k: v} )
                # new dict: result of applying changes to old (inlcudes non changed values)
                new = dict(old) #copy dict content, not dict pointer 
                new.update(changed)
                #Evaluate trigger expression
                eval_dict = dict(DEFAULT_EVALDICT)
                eval_dict.update({
                    'obj': obj, #allows object.notation
                    'old': old,
                    'changed': changed, 
                    'new': new, 
                    'inserting': is_ins, 'creating': is_ins,
                    'updating': not is_ins, 'writing': not is_ins,
                    })
                #rint '..old     = ', eval_dict['old']  #deepdebug
                #rint '..changed = ', eval_dict['changed']  #deepdebug
                #try: ...Leaving eval errors unhandled...
                ok = safe_eval(action.trg_evalexpr.replace('\n', ' '), {}, eval_dict) 
                if ok:
    #                if action.trg_evalexpr_debug:
                    _logger.info('Activated rule trigger %s on record id %d.' % (action.name, obj.id) )
                    _logger.info('CHG: %s' % str(changed))
                    _logger.info('NEW: %s' % str(new))
                    _logger.info('OLD: %s' % str(old))
    #                else:
    #                    _logger.debug('Activated rule trigger %s on record id %d.' % (action.name, obj.id) )
                    #rint '==>> RULE Activated:', action.name, action.trg_evalexpr, '====' #deepdebug
                    #rint ' old = ', eval_dict['old'] #deepdebug
                    #rint ' new = ', eval_dict['new'] #deepdebug
                    #rint '  changed = ', eval_dict['changed']  #deepdebug
                    #TODO?
                    #if not isinstance(ok, bool):
                    #    _logger.warn('Rule trigger %s on record id %d did not return a boolean.' % (action.name, obj.id) )
                #except (ValueError, KeyError, TypeError):
                #    _logger.error('Failed rule trigger %s on record id %d.' % (action.name, obj.id) )
                #    ok = False
                #rint '(eval rule %r = %s, record id %d.' % (ok, action.name, obj.id)
        return ok

    #Action able to send e-mails using email_template; 
    #Bonus: messages sent are recorded in the communication history.
    def do_action(self, cr, uid, action, model_obj, obj, context=None):
        super(base_action_rule, self).do_action(cr, uid, action, model_obj, obj, context=context)
        if action.email_template_id:
            _logger.info('...sending mail %s.' % (action.email_template_id.name) )
            mail_template = self.pool.get('email.template')
            mail_message = self.pool.get('mail.message')
            msg_id = mail_template.send_mail(cr, uid, action.email_template_id, obj.id, context=context)
            #mail_template does not set the e-mail date by itself!
            mail_message.write(cr, uid, [msg_id], {'date': time.strftime('%Y-%m-%d %H:%M:%S'),}, context=context)
            #send immediatly, if the option is checked
            if action.email_template_force:
                mail_message.send(cr, uid, [msg_id], context=context)
        return True

base_action_rule()

#Force register hooks on user Login, to not depend on scheduler
#BUG https://bugs.launchpad.net/openobject-addons/+bug/944197
class users(osv.osv):
    _inherit = "res.users"
    
    def login(self, db, login, password):
        #Perform login
        user_id = super(users, self).login(db, login, password)
        #Register hooks 
        cr = pooler.get_db(db).cursor()
        rule_pool = pooler.get_pool(db).get('base.action.rule')
        rule_ids = rule_pool.search(cr, 1, [])
        rule_pool._register_hook(cr, 1, rule_ids)
        rule_ids = None 
        cr.close()
        #End
        return user_id
users()
